/*
  cbess.c

  This program contains several functions for calculating bessel functions
  with complex arguments
*/

#include <math.h>

#define ACC 40.0
#define BIGNO 1.0e10
#define BIGNI 1.0e-10

#define maxiter  1000       /*Max. number of iterations for modified bessels*/
#define GAMMA 0.577215664901532860606512

typedef struct DCOMPLEX {double r,i;} dcomplex;

double Cabs();              /*Externally defined functions for Complex Operations*/
dcomplex Cmul();            /*These routines come from Numerical Recipes in C*/
dcomplex Cadd();
dcomplex Csub();
dcomplex Cdiv();
dcomplex Complex();
dcomplex RCmul();
dcomplex Csqrt();

dcomplex Csin();            /*Externally defined functions with complex arguments*/
dcomplex Ccos();
dcomplex Cexp();
dcomplex Clog();

dcomplex Cbessj0( ),Cbessj1( );   /*Internal Functions*/
dcomplex CBessJ0(), CBessJ1();
dcomplex CBessI0(), CBessI1();
dcomplex CBessK0(),CBessK1();

/*****************************************************
  SPHERICAL BESSEL FUNCTION OF THE FIRST KIND
  Adapted for complex arguments from 
  Numerical Recipes in C.
*****************************************************/
dcomplex SphBesselJ(n,x)
int n;
dcomplex x;
{
   int j,jsum,m;
   dcomplex ax,tox,bj,bjm,bjp,ans,sum;
   void nrerror();
   
   if (n==0) 
     return Cbessj0(x); 
   
   if (n==1) 
     return Cbessj1(x); 
   
   if (n > 1)
     { ax=x;
       if ((ax.r==0.) && (ax.i==0.))
	 return Complex(0.0,0.0);
       else if (Cabs(ax) > (double) n) {
	 tox=Cdiv(Complex(2.,0.),ax);
	 bjm=Cbessj0(ax);
	 bj=Cbessj1(ax);
	 for (j=1;j<n;j++) {
	   bjp=Csub(Cmul(RCmul((double)(j+.5),tox),bj),bjm);
	   bjm=bj;
	   bj=bjp;
	 }
	 ans=bj;
       } else {
	 tox=Cdiv(Complex(2.,0.),ax);
	 m=2*((n+(int)sqrt(ACC*n))/2);
	 bjp=ans=Complex(0.0,0.0);
	 bj=Complex(1.0,2.0);
	 for (j=m;j>0;j--) {
	   bjm=Csub(Cmul(RCmul((double)(j+.5),tox),bj),bjp);
	   bjp=bj;
	   bj=bjm;
	   if (Cabs(bj) > BIGNO) {
	     bj = RCmul(BIGNI,bj);
	     bjp = RCmul(BIGNI,bjp);
	     ans = RCmul(BIGNI,ans);
	   } 
	   if (j == n) ans=bjp;
	 }
	 sum = Cdiv(SphBesselJ(0,ax),bj);
	 ans=Cmul(ans,sum);
       }
       return  ans;
     }
}
dcomplex Cbessj0(x) dcomplex x;
{  return Cdiv( Csin(x), x );}
dcomplex Cbessj1(x) dcomplex x;
{  dcomplex temp;
   temp = Cdiv( Csin(x), Cmul(x,x) );
   temp = Csub(temp, Cdiv( Ccos(x), x ) );
   return temp; }


/*********************************************************
  DERIVATIVE SPHERICAL BESSEL FUNCTION OF THE FIRST KIND
*********************************************************/
dcomplex DSphBesselJ( l, x )
dcomplex x;
int l;
{
  dcomplex j;
  
  if( l==0 )
    {
      j = RCmul( -1.0, SphBesselJ( 1, x ) );
      return j;
    }

  j = RCmul( 1.0/(double)(2*l+1), Csub( RCmul((double)l,SphBesselJ(l-1,x)), RCmul((double)(l+1),SphBesselJ(l+1,x))));
  return j;
}


/*****************************************************
  SPHERICAL HANKEL FUNCTION OF THE FIRST KIND
  Adapted for complex arguments from
  Numerical Recipes in C.
*****************************************************/
dcomplex SphBesselH( l, x )
dcomplex x;
int l;
{
  dcomplex h, h1, h2, y, c;
  int i;

  c = Complex( 0.0, 1.0 );
  h2 = Cdiv( Cexp(Cmul(c,x)),Cmul(c,x) );
  h1 = Cmul( Cdiv(Cexp(Cmul(c,x)),x), Cadd( Complex(-1.0,0.0),Cdiv(Complex(0.0,-1.0),x)));

  if( l==0 )  return h2;
  if( l==1 )  return h1;
  for( i=1; i<l; i++ )
    {
      y = Cdiv( h1, x );
      h.r = (2*i+1) * y.r - h2.r;
      h.i = (2*i+1) * y.i - h2.i;
      h2.r = h1.r;
      h2.i = h1.i;
      h1.r = h.r;
      h1.i = h.i;
    }
  return h;
}


/*********************************************************
  DERIVATIVE SPHERICAL HANKEL FUNCTION OF THE FIRST KIND
*********************************************************/
dcomplex DSphBesselH( l, x )
dcomplex x;
int l;
{
  dcomplex h;
  
  if( l==0 )
    {
      h = RCmul( -1.0, SphBesselH( 1, x ) );
      return h;
    }

  h = RCmul( 1.0/(double)(2*l+1), Csub( RCmul((double)l,SphBesselH(l-1,x)), RCmul((double)(l+1),SphBesselH(l+1,x))));
  return h;
}


/*******************************************************/
/* Bessel Function of the first kind calculated using  */
/* backwards recurence and the normalization condition */
/* J0(z) + 2 J2(z) + 2 J4(z) + ... = 1                 */
/* This routine has been adapted for complex arguments */
/* from a routine given in Numerical Recipes.          */
/* Although the Numerical Recipes routine works for    */
/* negative real arguments, this routine will return   */
/* incorrect answers.                                  */
/*******************************************************/
/* NOT COMPLETED                                       */
/*******************************************************/
dcomplex CBessJ(n,z)
int n;
dcomplex z;
{
  int j,jsum,m;
  dcomplex bj,bjm,bjp,sum,tox,ans;
  dcomplex bessj0(),bessj1();
  void nrerror();

  if( n==0 )  return CBessJ0(z);
  if( n==1 )  return CBessJ1(z);

  if(  Cabs(z) == 0.0 )  return Complex(0.0,0.0);
  else if( Cabs(z) > (double)n )  {
    tox=Cdiv( Complex(2.,0.), z );
    bjm=CBessJ0(z);
    bj=CBessJ1(z);
    for (j=1;j<n;j++) {
      bjp=Csub(RCmul((double)j,Cmul(tox,bj)),bjm);
      bjm.r=bj.r;  bjm.i=bj.i;
      bj.r=bjp.r; bj.i=bjp.i;
    }
    ans.r=bj.r;  ans.i=bj.i;
  } else {
    tox=Cdiv(Complex(2.0,0.0),z);
    m=2*((n+(int) sqrt(ACC*n))/2);
    jsum=0;
    bjp=Complex(0.,0.);
    ans=Complex(0.,0.);
    sum=Complex(0.,0.);
    bj=Complex(1.0,0.);
    for (j=m;j>0;j--) {
      bjm=Csub(RCmul((double)j,Cmul(tox,bj)),bjp);
      bjp.r=bj.r;  bjp.i=bj.i;
      bj.r=bjm.r;  bj.i=bjm.i;
      if( Cabs(bj) > BIGNO) {
	bj = RCmul(BIGNI,bj);
	bjp = RCmul(BIGNI,bjp);
	ans = RCmul(BIGNI,ans);
	sum = RCmul(BIGNI,sum);
      }
      if( jsum ) sum = Cadd(sum,bj);
      jsum=!jsum;
      if (j == n) {ans.r=bjp.r;  ans.i=bjp.i;}
    }
    sum=Csub(RCmul(2.0,sum),bj);
    ans = Cdiv(ans,sum);
  }
  return ans;
}

dcomplex CBessJ0(z)
dcomplex z;
{
  printf( "CBessJ0 not implemented yet.\n" );
  return z;
}

dcomplex CBessJ1(z)
dcomplex z;
{
  printf( "CBessJ1 not implemented yet.\n" );
  return z;
}

/*******************************************************/
/* Modified Bessel Function I_n(z)                     */
/* This routine has been adapted for complex arguments */
/* from a routine given in Numerical Recipes.          */
/* Although the Numerical Recipes routine works for    */
/* negative real arguments, this routine will return   */
/* incorrect answers.                                  */
/*******************************************************/
dcomplex CBessI(n,z)
int n;
dcomplex z;
{
  int j;
  dcomplex bi,bim,bip,tox,ans;
  dcomplex bessi0();
  void nrerror();

  if( n==0 )  return CBessI0(z);
  if( n==1 )  return CBessI1(z);

  if( Cabs(z) == 0.0 )  return Complex(0.,0.);
  else {
    tox=Cdiv(Complex(2.0,0.),z);
    bip=Complex(0.,0.);
    ans=Complex(0.,0.);
    bi=Complex(1.0,0.);
    for (j=2*(n+(int) sqrt(ACC*n));j>0;j--) {
      bim=Cadd(bip,RCmul((double)j,Cmul(tox,bi)));
      bip.r=bi.r;  bip.i=bi.i;
      bi.r=bim.r;  bi.i=bim.i;
	if( Cabs(bi) > BIGNO) {
	  ans = RCmul(BIGNI,ans);
	  bi = RCmul(BIGNI,bi);
	  bip = RCmul(BIGNI,bip);
	}
      if (j == n)  ans=bip;
    }
    ans = Cmul(ans,Cdiv(CBessI0(z),bi));
    return  ans;
  }
}

dcomplex CBessI0(z)
dcomplex z;
{
  dcomplex ans,y;

  if (Cabs(z) < 3.75) {
    y=RCmul(1./3.75,z);
    y=Cmul(y,y);
    ans=Cadd(Complex(1.,0.),Cmul(y,Cadd(Complex(3.5156229,0.),
			    Cmul(y,Cadd(Complex(3.0899424,0.),
			    Cmul(y,Cadd(Complex(1.2067492,0.),
			    Cmul(y,Cadd(Complex(0.2659732,0.),
			    Cmul(y,Cadd(Complex(0.360768e-1,0.),
			    RCmul(0.45813e-2,y))))))))))));
  } else {
    y=Cdiv(Complex(3.75,0.),z);
    ans=Cadd(Complex(0.39894228,0.),
	                    Cmul(y,Cadd(Complex(0.1328592e-1,0.),
			    Cmul(y,Cadd(Complex(0.225319e-2,0.),
			    Cmul(y,Cadd(Complex(-0.157565e-2,0.),
			    Cmul(y,Cadd(Complex(0.916281e-2,0.),
			    Cmul(y,Cadd(Complex(-0.2057706e-1,0.),
			    Cmul(y,Cadd(Complex(0.2635537e-1,0.),
			    Cmul(y,Cadd(Complex(-0.1647633e-1,0.),
			    RCmul(0.392377e-2,y))))))))))))))));
    y = Cexp(z);
    y = Cdiv( y, Csqrt(z) );
    ans=Cmul(ans,y);
  }
  return ans;
}

dcomplex CBessI1(z)
dcomplex z;
{
  dcomplex ans,y;

  if (Cabs(z) < 3.75) {
    y=RCmul(1./3.75,z);
    y=Cmul(y,y);
    ans=Cmul(z,Cadd(Complex(0.5,0.),
	Cmul(y,Cadd(Complex(0.87890594,0.),
	Cmul(y,Cadd(Complex(0.51498869,0.),
	Cmul(y,Cadd(Complex(0.15084934,0.),
	Cmul(y,Cadd(Complex(0.2658733e-1,0.),
	Cmul(y,Cadd(Complex(0.301532e-2,0.),
	RCmul(0.32411e-3,y)))))))))))));
  } else {
    y=Cdiv(Complex(3.75,0.),z);
    ans=Cadd(Complex(0.2282967e-1,0.),
	                    Cmul(y,Cadd(Complex(-0.2895312e-1,0.),
			    Cmul(y,Cadd(Complex(0.1787654e-1,0.),
			    RCmul(-0.420059e-2,y))))));
    ans=Cadd(Complex(0.39894228,0.),
	                    Cmul(y,Cadd(Complex(-0.3988024e-1,0.),
			    Cmul(y,Cadd(Complex(-0.362018e-2,0.),
			    Cmul(y,Cadd(Complex(0.163801e-2,0.),
			    Cmul(y,Cadd(Complex(-0.1031555e-1,0.),
			    Cmul(ans,y))))))))));
    y = Cexp(z);
    y = Cdiv( y, Csqrt(z) );
    ans=Cmul(ans,y);
  }
  return ans;
}


/***************************************************/
/*Derivative of the modified bessel function I_n(z)*/
/***************************************************/
dcomplex CDBessI(n,z)
int n;
dcomplex z;
{
  dcomplex u;

  u = Cadd( CBessI(n+1,z), RCmul(n,Cdiv(CBessI(n,z),z)) );
  return u;
}



/*******************************************************/
/* Modified Bessel Function K_n(z)                     */
/* This routine has been adapted for complex arguments */
/* from a routine given in Numerical Recipes.          */
/* Although the Numerical Recipes routine works for    */
/* negative real arguments, this routine will return   */
/* incorrect answers.                                  */
/*******************************************************/
dcomplex CBessK(n,z)
int n;
dcomplex z;
{
  int j;
  dcomplex bk,bkm,bkp,tox;
  void nrerror();

  if( n==0 )  return CBessK0(z);
  if( n==1 )  return CBessK1(z);

  tox=Cdiv( Complex(2.0,0.), z );
  bkm=CBessK0(z);
  bk=CBessK1(z);
  for (j=1;j<n;j++) {
    bkp=Cadd(bkm,RCmul((double)j,Cmul(tox,bk)));
    bkm=bk;
    bk=bkp;
  }
  return bk;
}

dcomplex CBessK0(z)
dcomplex z;
{
  dcomplex y,ans;
  dcomplex CBessI0();

  if (Cabs(z) <= 2.0) {
    y=RCmul(0.25, Cmul(z,z) );
    ans=Cadd(Complex(-0.57721566,0.),Cmul(y,Cadd(Complex(0.42278420,0.),
			    Cmul(y,Cadd(Complex(0.23069756,0.),
			    Cmul(y,Cadd(Complex(0.3488590e-1,0.),
			    Cmul(y,Cadd(Complex(0.262698e-2,0.),
			    Cmul(y,Cadd(Complex(0.10750e-3,0.),
			    RCmul(0.74e-5,y))))))))))));
    ans = Cadd( ans, RCmul( -1., Cmul(Clog(RCmul(.5,z)),CBessI0(z))) );
  } else {
    y = Cdiv( Complex(2.0,0.0), z );
    ans=Cadd(Complex(1.25331414,0.),Cmul(y,Cadd(Complex(-0.7832358e-1,0.),
			    Cmul(y,Cadd(Complex(0.2189568e-1,0.),
			    Cmul(y,Cadd(Complex(-0.1062446e-1,0.),
			    Cmul(y,Cadd(Complex(0.587872e-2,0.),
			    Cmul(y,Cadd(Complex(-0.251540e-2,0.),
			    RCmul(0.53208e-3,y))))))))))));
    y = Cexp( RCmul(-1.,z) );
    y = Cdiv( y, Csqrt(z) );
    ans = Cmul( ans, y );
  }
  return ans;
}

dcomplex CBessK1(z)
dcomplex z;
{
  dcomplex y,ans;

  if (Cabs(z) <= 2.0) {
    y=RCmul(0.25, Cmul(z,z) );
    ans=Cmul(Cdiv(Complex(1.0,0.),z),Cadd(Complex(1.0,0.),
			    Cmul(y,Cadd(Complex(0.15443144,0.),
			    Cmul(y,Cadd(Complex(-0.67278579,0.),
			    Cmul(y,Cadd(Complex(-0.18156897,0.),
			    Cmul(y,Cadd(Complex(-0.1919402e-1,0.),
			    Cmul(y,Cadd(Complex(-0.110404e-2,0.),
			    RCmul(-0.4686e-4,y)))))))))))));
    ans = Cadd( ans, Cmul(Clog(RCmul(.5,z)),CBessI1(z)) );
  } else {
    y = Cdiv( Complex(2.0,0.0), z );
    ans=Cadd(Complex(1.25331414,0.),Cmul(y,Cadd(Complex(0.23498619,0.),
			    Cmul(y,Cadd(Complex(-0.3655620e-1,0.),
			    Cmul(y,Cadd(Complex(0.1504268e-1,0.),
			    Cmul(y,Cadd(Complex(-0.780353e-2,0.),
			    Cmul(y,Cadd(Complex(0.325614e-2,0.),
			    RCmul(-0.68245e-3,y))))))))))));
    y = Cexp( RCmul(-1.,z) );
    y = Cdiv( y, Csqrt(z) );
    ans = Cmul( ans, y );
  }
  return ans;
}


/***************************************************/
/*Derivative of the modified bessel function K_n(z)*/
/***************************************************/
dcomplex CDBessK(n,z)
int n;
dcomplex z;
{
  dcomplex u;

  u = Csub( RCmul(n,Cdiv(CBessK(n,z),z)), CBessK(n+1,z) );
  return u;
}



#undef ACC
#undef BIGNO
#undef BIGNI
