/*****************************************************
  SOLUTIONS TO THE HELMHOLTZ EQUATION

COPYRIGHT:
     Copyright 1994 by David Boas (boas@sol1.lrsm.upenn.edu)

     Permission to use, copy, modify, and distribute this
     software and its documentation for any purpose and without
     fee is hereby granted, provided that the above copyright
     notice appear in all copies.  The University of Pennsylvania
     makes no representations about the suitability of this
     software for any purpose.  It is provided "as is" without
     express or implied warranty.

DISCLAIMER:
     UNDER NO CIRCUMSTANCES WILL THE AUTHOR OF THIS SOFTWARE OR THE
     UNIVERSITY OF PENNSYLVANIA BE RESPONSIBLE FOR ANY DIRECT OR
     INCIDENTAL DAMAGE ARISING FROM THE USE OF THIS SOFTWARE AND ITS
     DOCUMENTATION. THE SOFTWARE HEREIN IS PROVIDED "AS IS" WITH NO
     IMPLIED OBLIGATION TO PROVIDE SUPPORT, UPDATES, OR MODIFICATIONS.
*****************************************************/

#include "p1.h"


/*****************************************************
  HANDLE A FINITE SIZED DETECTOR
*****************************************************/
dcomplex FSD( k, r, theta, z_s, a, r_d, step, func )
dcomplex k;
double r, theta, z_s, a, r_d, step;
int func;
{
  dcomplex u, u1;
  double xc, zc, x, z, xp, zp;

  u = Complex(0.,0.);
  xc = r*sin(theta);
  zc = r*cos(theta);

  for( x=-r_d; x<=r_d; x+=step )
    for( z=-r_d; z<=r_d; z+=step )
      {
	r = sqrt( x*x + z*z );
	if( r<=r_d )
	  {
	    xp = x+xc;
	    zp = z+zc;
	    r = sqrt( xp*xp + zp*zp );
	    if( func==0 )
	      {
		u1 = Inc( k, r );
	      }
	    else
	      {
		theta = acos( zp/r );
		u1 = ScSphAbs2( k, r, theta, z_s, a );
	      }
	    u = Cadd(u,u1);
	  }
      }
  return u;
}

/*****************************************************
  CALULATE THE INCIDENT WAVE
******************************************************/
dcomplex Inc( k, r )
dcomplex k;
double r;
{
  dcomplex u_inc, x;

  x=Cmul(k,Complex(0.,r));
  u_inc = RCmul( 1./(4.*pi*r), Cexp(x) );

  return u_inc;
}

/*****************************************************
  CALCULATE THE WAVE SCATTERED FROM A PERFECT ABSORBER
  SPHERE USING THE ZERO BOUNDARY CONDITION
******************************************************/
dcomplex ScSphAbs( k, r, theta, z_s, a )
dcomplex k;
double r, theta, z_s, a;
{
  dcomplex u_sc, u_add, x, y, z;
  int l;

  x = RCmul(a, k);
  y = RCmul(z_s, k);
  z = RCmul(r, k);

  l=1;
  u_sc = Cdiv( Cmul( Complex(0.0,-1.0), Cmul( k, Cmul( SphBesselJ(0,x), Cmul( SphBesselH(0,y), Cmul( SphHarmonic(0,3.14159), Cmul( SphBesselH(0,z), SphHarmonic(0,theta))))))), SphBesselH(0,x) );
  u_add = Cdiv( Cmul( Complex(0.0,-1.0), Cmul( k, Cmul( SphBesselJ(1,x), Cmul( SphBesselH(1,y), Cmul( SphHarmonic(1,3.14159), Cmul( SphBesselH(1,z), SphHarmonic(1,theta))))))), SphBesselH(1,x) );
  while( Cabs(u_add)/Cabs(u_sc) > tollerence )
    {
      l++;
      u_sc = Cadd( u_sc, u_add );
      u_add = Cdiv( Cmul( Complex(0.0,-1.0), Cmul( k, Cmul( SphBesselJ(l,x), Cmul( SphBesselH(l,y), Cmul( SphHarmonic(l,3.14159), Cmul( SphBesselH(l,z), SphHarmonic(l,theta))))))), SphBesselH(l,x) );
      
    }

  return u_sc;
}

/*****************************************************
  CALCULATE THE WAVE SCATTERED FROM A PERFECT ABSORBER
  SPHERE USING THE LIMITING FORM OF ScDisp
******************************************************/
dcomplex ScSphAbs2( k, r, theta, z_s, a )
dcomplex k;
double r, theta, z_s, a;
{
  dcomplex u_sc, u_add, x, y, z, dk, temp;
  int l;
  double foo;

  x = RCmul(a, k);
  y = RCmul(z_s, k);
  z = RCmul(r, k);

  dk = RCmul(-sqrt(3.0)*Do,k);

  l=1;
  temp = ScSphAbs2_coeff_1( 0, k, x, y, dk );
  u_sc = Cmul( temp, Cmul( SphBesselH(0,z), SphHarmonic(0,theta)));
  temp = ScSphAbs2_coeff_1( 1, k, x, y, dk );
  u_add = Cmul( temp,Cmul( SphBesselH(1,z),SphHarmonic(1,theta)));
  while( Cabs(u_add)/Cabs(u_sc) > tollerence )
    {
      l++;
      u_sc = Cadd( u_sc, u_add );
      temp = ScSphAbs2_coeff_1( l, k, x, y, dk );
      u_add = Cmul( temp,Cmul( SphBesselH(l,z),SphHarmonic(l,theta)));
    }

  return u_sc;
}
dcomplex ScSphAbs2_coeff( l, k, z_s, a )
dcomplex k;
double z_s, a;
int l;
{
  dcomplex x, y, dk;
  x = RCmul(a, k);
  y = RCmul(z_s, k);
  dk = RCmul(-2.*sqrt(3)*Do,k);
  return ScSphAbs2_coeff_1( l, k, x, y, dk );
}
dcomplex ScSphAbs2_coeff_1( l, k, x, y, dk )
dcomplex k, x, y, dk;
int l;
{
  dcomplex jtemp, htemp;

  jtemp = Cadd(SphBesselJ(l,x), Cmul(dk, DSphBesselJ(l,x)));
  htemp = Cadd(SphBesselH(l,x), Cmul(dk, DSphBesselH(l,x)));
  return Cmul( Cdiv(jtemp,htemp),Cmul( Complex(0.0,-1.0),Cmul( k,Cmul( SphBesselH(l,y),SphHarmonic(l,pi)))));
}

/*********************************************************************
  CALCULATE THE WAVE SCATTERED FROM A DISPERSIVE SPHERE
*********************************************************************/
dcomplex ScSphDisp( ko, ki, r, theta, z_s, a )
dcomplex ko, ki;
double r, theta, z_s, a;
{
  dcomplex u_sc, u_add, xo, xi, y, z, dxo, dxi, temp;
  dcomplex jo, ji, ho;
  int l;

  xo = RCmul(a, ko);
  xi = RCmul(a, ki);
  y = RCmul(z_s, ko);
  z = RCmul(r, ko);

  dxo = RCmul(Do, xo);
  dxi = RCmul(Di[obj_num], xi);

  l=1;
  temp = ScSphDisp_coeff_1(0,ko,xo,xi,dxo,dxi,y);
  u_sc = Cmul( temp, Cmul( SphBesselH(0,z), SphHarmonic(0,theta)));
  temp = ScSphDisp_coeff_1(1,ko,xo,xi,dxo,dxi,y);
  u_add = Cmul( temp, Cmul( SphBesselH(1,z), SphHarmonic(1,theta)));
  while( Cabs(u_add)/Cabs(u_sc) > tollerence )
    {
      l++;
      u_sc = Cadd( u_sc, u_add );
      temp = ScSphDisp_coeff_1(l,ko,xo,xi,dxo,dxi,y);
      u_add = Cmul( temp, Cmul( SphBesselH(l,z), SphHarmonic(l,theta)));
    }
  return u_sc;
}
dcomplex ScSphDisp_coeff(l,ko,ki,z_s,a)
dcomplex ko,ki;
double z_s,a;
int l;
{
  dcomplex xo,xi,y,dxo,dxi;

  xo = RCmul(a, ko);
  xi = RCmul(a, ki);
  y = RCmul(z_s, ko);
  dxo = RCmul(Do, xo);
  dxi = RCmul(Di[obj_num], xi);
  return ScSphDisp_coeff_1(l,ko,xo,xi,dxo,dxi,y);
}
dcomplex ScSphDisp_coeff_1(l,k,xo,xi,dxo,dxi,y)
dcomplex k,xo,xi,dxo,dxi,y;
int l;
{ 
  dcomplex jo,ji,ho,temp;

  jo = SphBesselJ(l,xo);
  ji = SphBesselJ(l,xi);
  ho = SphBesselH(l,xo);
  temp = Cadd(Cdiv(jo,ho),Cmul(Cdiv(ji,ho),Cdiv(Csub(Cmul(ho,DSphBesselJ(l,xo)),Cmul(DSphBesselH(l,xo),jo)),Csub(Cmul(DSphBesselH(l,xo),ji),Cmul(Cdiv(dxi,dxo),Cmul(ho,DSphBesselJ(l,xi)))))));
  return Cmul(temp,Cmul(Complex(0.0,-1.0),Cmul(k,Cmul(SphBesselH(l,y),SphHarmonic(l,pi)))));
}


/*******************************************************************
  CALCULATE THE WAVE SCATTERED FROM A PERFECT CYLINDRICAL ABSORBER
  OF INFINITE LENGTH, USING THE ZERO BOUNDARY CONDITION
*******************************************************************/
dcomplex gk2;
double gr,gtheta,gz,gr_s,ga;
double amp_o;
dcomplex quad();

dcomplex ScCylAbs( k2, r, theta, z, r_s, a )
dcomplex k2;
double r, theta, z, r_s, a;
{
  dcomplex u_sc;
  dcomplex Q_ScCyl();

  gk2 = k2;
  gr = r;
  gtheta = theta;
  gz = z;
  gr_s = r_s;
  ga = a;

  amp_o=0.;
  u_sc = quad( Q_ScCyl, 1.e-8, 100., 1e-3 );

  return RCmul( -1./(2.*pi*pi), u_sc );
}

dcomplex Q_ScCyl( p )
double p;
{
  dcomplex kappa, xa, xs, xr, u, u_add;
  int n;

  kappa = Csqrt( Cadd( Complex(p*p,0.), gk2 ) );
  xa = RCmul( ga, kappa );
  xs = RCmul( gr_s, kappa );
  xr = RCmul( gr, kappa );

  u = Cdiv( Cmul( CBessI(0,xa), Cmul( CBessK(0,xs), CBessK(0,xr) ) ),
		CBessK(0,xa) );
  if( amp_o==0. )  amp_o = Cabs(u);

  n=1;
  u_add = Cdiv( Cmul( CBessI(n,xa), Cmul( CBessK(n,xs), CBessK(n,xr) ) ),
		CBessK(n,xa) );
  while( Cabs(u_add)/amp_o > tollerence )
    {
      u = Cadd( u, RCmul( 2.*cos((double)n * gtheta), u_add ) );
      n++;
      u_add = Cdiv( Cmul( CBessI(n,xa), Cmul( CBessK(n,xs), CBessK(n,xr) ) ),
	       CBessK(n,xa) );
    }
  u = RCmul( cos(p*gz), u );

  return u;
}
  

/*******************************************************************
  CALCULATE THE WAVE SCATTERED FROM A PERFECT CYLINDRICAL ABSORBER
  OF INFINITE LENGTH, USING THE LIMITING FORM OF THE DISPERSIVE
  CYLINDER SOLUTION AS mu_a,in goes to infinity
*******************************************************************/
dcomplex gk2;
double gr,gtheta,gz,gr_s,ga;
double amp_o;
dcomplex quad();

dcomplex ScCylAbs2( k2, r, theta, z, r_s, a )
dcomplex k2;
double r, theta, z, r_s, a;
{
  dcomplex u_sc;
  dcomplex Q_ScCyl2();

  gk2 = k2;
  gr = r;
  gtheta = theta;
  gz = z;
  gr_s = r_s;
  ga = a;

  amp_o=0.;
  u_sc = quad( Q_ScCyl2, 1.e-8, 100., 1e-3 );

  return RCmul( -1./(2.*pi*pi), u_sc );
}

dcomplex Q_ScCyl2( p )
double p;
{
  dcomplex kappa, xa, xs, xr, u, u_add;
  dcomplex foo, foo1, foo2;
  int n;

  kappa = Csqrt( Cadd( Complex(p*p,0.), gk2 ) );
  xa = RCmul( ga, kappa );
  xs = RCmul( gr_s, kappa );
  xr = RCmul( gr, kappa );
  foo = RCmul( -sqrt(3.0)*Do, kappa );

  foo1 = Csub( CBessI(0,xa), Cmul(foo,CDBessI(0,xa)) );
  foo2 = Csub( CBessK(0,xa), Cmul(foo,CDBessK(0,xa)) );
  u = Cmul( CBessK(0,xs), Cmul(CBessK(0,xr),Cdiv(foo1,foo2)) );
  if( amp_o==0. )  amp_o = Cabs(u);

  n=1;
  foo1 = Csub( CBessI(n,xa), Cmul(foo,CDBessI(n,xa)) );
  foo2 = Csub( CBessK(n,xa), Cmul(foo,CDBessK(n,xa)) );
  u_add = Cmul( CBessK(n,xs), Cmul(CBessK(n,xr),Cdiv(foo1,foo2)) );
  while( Cabs(u_add)/amp_o > tollerence )
    {
      u = Cadd( u, RCmul( 2.*cos((double)n * gtheta), u_add ) );
      n++;
      foo1 = Csub( CBessI(n,xa), Cmul(foo,CDBessI(n,xa)) );
      foo2 = Csub( CBessK(n,xa), Cmul(foo,CDBessK(n,xa)) );
      u_add = Cmul( CBessK(n,xs), Cmul(CBessK(n,xr),Cdiv(foo1,foo2)) );
    }
  u = RCmul( cos(p*gz), u );

  return u;
}


/*******************************************************************
  Calculate the wave scattered from a dispersive cylinder
  of infinite length.
*******************************************************************/
dcomplex gko2, gki2;
double gr,gtheta,gz,gr_s,ga;
double amp_o;
dcomplex quad();

dcomplex ScCylDisp( ko2, ki2, r, theta, z, r_s, a )
dcomplex ko2, ki2;
double r, theta, z, r_s, a;
{
  dcomplex u_sc;
  dcomplex Q_ScCylDisp();

  gko2 = ko2;
  gki2 = ki2;
  gr = r;
  gtheta = theta;
  gz = z;
  gr_s = r_s;
  ga = a;

  amp_o=0.;
  u_sc = quad( Q_ScCylDisp, 1.e-8, 100., 1e-3 );

  return RCmul( -1./(2.*pi*pi), u_sc );
}

dcomplex Q_ScCylDisp( p )
double p;
{
  dcomplex kappa_out, kappa_in, xa, ya, xs, xr, u, u_add;
  dcomplex foo, foo1, foo2;
  int n;

  kappa_out = Csqrt( Cadd( Complex(p*p,0.), gko2 ) );
  kappa_in = Csqrt( Cadd( Complex(p*p,0.), gki2 ) );
  xa = RCmul( ga, kappa_out );
  ya = RCmul( ga, kappa_in );
  xs = RCmul( gr_s, kappa_out );
  xr = RCmul( gr, kappa_out );
  foo = RCmul( -Di[obj_num]/Do, Cdiv(kappa_in,kappa_out) );

  foo1 = Csub( Cmul(CBessI(0,ya),CDBessI(0,xa)), Cmul(foo,Cmul(CDBessI(0,ya),CBessI(0,xa))) );
  foo2 = Csub( Cmul(CBessI(0,ya),CDBessK(0,xa)), Cmul(foo,Cmul(CDBessI(0,ya),CBessK(0,xa))) );
  u = Cmul( CBessK(0,xs), Cmul(CBessK(0,xr),Cdiv(foo1,foo2)) );
  if( amp_o==0. )  amp_o = Cabs(u);

  n=1;
  foo1 = Csub( Cmul(CBessI(n,ya),CDBessI(n,xa)), Cmul(foo,Cmul(CDBessI(n,ya),CBessI(n,xa))) );
  foo2 = Csub( Cmul(CBessI(n,ya),CDBessK(n,xa)), Cmul(foo,Cmul(CDBessI(n,ya),CBessK(n,xa))) );
  u_add = Cmul( CBessK(n,xs), Cmul(CBessK(n,xr),Cdiv(foo1,foo2)) );
  while( Cabs(u_add)/amp_o > tollerence )
    {
      u = Cadd( u, RCmul( 2.*cos((double)n * gtheta), u_add ) );
      n++;
      foo1 = Csub( Cmul(CBessI(n,ya),CDBessI(n,xa)), Cmul(foo,Cmul(CDBessI(n,ya),CBessI(n,xa))) );
      foo2 = Csub( Cmul(CBessI(n,ya),CDBessK(n,xa)), Cmul(foo,Cmul(CDBessI(n,ya),CBessK(n,xa))) );
      u_add = Cmul( CBessK(n,xs), Cmul(CBessK(n,xr),Cdiv(foo1,foo2)) );
    }
  u = RCmul( cos(p*gz), u );

  return u;
}
  

