/*
  P1.C

  Solve for the propagation of a DPDW in a homogeneous medium except
  for the presence of a localized inhomogeneity.  The localized
  inhomogeneity could be either a

         Sphere
	 Cylinder of infinite length

  The analytic result has been derived from the Helmholtz equation.
  The boundary conditions are outlined in HELMHOLTZ.C

COPYRIGHT:
     Copyright 1994 by David Boas (boas@sol1.lrsm.upenn.edu)

     Permission to use, copy, modify, and distribute this
     software and its documentation for any purpose and without
     fee is hereby granted, provided that the above copyright
     notice appear in all copies.  The University of Pennsylvania
     makes no representations about the suitability of this
     software for any purpose.  It is provided "as is" without
     express or implied warranty.

DISCLAIMER:
     UNDER NO CIRCUMSTANCES WILL THE AUTHOR OF THIS SOFTWARE OR THE
     UNIVERSITY OF PENNSYLVANIA BE RESPONSIBLE FOR ANY DIRECT OR
     INCIDENTAL DAMAGE ARISING FROM THE USE OF THIS SOFTWARE AND ITS
     DOCUMENTATION. THE SOFTWARE HEREIN IS PROVIDED "AS IS" WITH NO
     IMPLIED OBLIGATION TO PROVIDE SUPPORT, UPDATES, OR MODIFICATIONS.
*/

#include "p1.h"


main( argc, argv )
int argc;
char *argv[];
{
  dcomplex u, ko, ki[NOBJS], uinc, usc;
  VECTOR ro[NOBJS], rb[NPLANES];
  SOURCE rs[NSRCS],rd[NSRCS];
  double amp, ampsc, amp_o;
  double ph[100][100], phsc[100][100], phinc[100][100];
  double phase, phold, phscold, phincold;
  double us, ua, usi, uai[NOBJS], w;
  double Zb, s;
  double kmag, karg;
  double xmin, xmax, xstep;
  double ymin, ymax, ystep;
  double zmin, zmax, zstep;
  double a[NOBJS];
  double foo1, foo2;
  int i, j, k, nx, ny, nz, nsrcs, ndets, nobjs, nplanes, npts;
  FILE *fp_inphm, *fp_insdg, *fp_outpdw, *fp_outsc, *fp_outinc, *fp_outxyz, *fp;
  char filenm[128], filetemp[132];

 /*Check command line*/
  if( argc!=2 && argc!=5 ) {
    printf( "\n" );
    printf( "Usage: p1 obj_type\n" );
    printf( "       obj_type is either\n" );
    printf( "           s - Sphere\n" );
    printf( "           c - Cylinder\n" );
    printf( "Or\n" );
    printf( "       p1 obj_type phantom-input[.phm]\n");
    printf( "                   source-detector-input[.sdg]\n");
    printf( "                   output[.pdw .sc]\n" );
    printf( "File names should not include extensions.\n" );
    printf( "\n" );
    exit(1);
  }

  switch( argv[1][0] ) {
  case 's':
    obj_type = SPHERE;
    break;
  case 'c':
    obj_type = CYLINDER;
    break;
  default:
    printf( "Object type must be either\n" );
    printf( "           s - Sphere\n" );
    printf( "           c - Cylinder\n" );
    exit(1);
  }

  if( argc==5 ) {
    sprintf( filenm,  "%s.phm", argv[2] );
    fp_inphm  = fopen( filenm,  "r" );

    sprintf( filenm,  "%s.sdg", argv[3] );
    fp_insdg  = fopen( filenm,  "r" );

    sprintf( filenm, "%s.pdw", argv[4] );
    fp_outpdw = fopen( filenm, "w" );

    sprintf( filenm, "%s.sc", argv[4] );
    fp_outsc = fopen( filenm, "w" );

    sprintf( filenm, "%s.inc", argv[4] );
    fp_outinc = fopen( filenm, "w" );

    sprintf( filenm, "%s.pdw.fit", argv[4] );
    fp = fopen( filenm, "w" );

    fscanf( fp_inphm, "%d %d", &nobjs, &nplanes );
    fscanf( fp_insdg, "%d %d", &nsrcs, &ndets );
  }
  else {
    printf( "Number of sources, objects, and free-space planar boundaries:" );
    scanf( "%d %d %d", &nsrcs, &nobjs, &nplanes );
    ndets = 1;
  }

  if( nsrcs>=NSRCS || ndets>=NSRCS || nobjs>=NOBJS || nplanes>=NPLANES ) {
    printf( "Must change maximum number of allowed\n" );
    printf( "sources, detectors, objects, or planes\n" );
    exit(1);
  }

/*HOMOGENEOUS MEDIUM INFORMATION*/
  if( argc==5 ) {
    fscanf( fp_inphm, "%lf %lf %lf", &us, &ua, &v );  v=3.0e10/v;
    fscanf( fp_insdg, "%lf", &w );
    printf( "     mu_s' = %lf\n", us );
    printf( "      mu_a = %lf\n", ua );
    printf( "         v = %lf\n", v );
    printf( "Mod. Freq. = %lf MHz\n", w );
    w = 2.*pi*w*1.0e6;
  }
  else {
    printf( "HOMOGENEOUS MEDIUM:\n mu_s' and mu_a (units of inverse centimeters):");
    scanf( "%lf %lf", &us, &ua);
    printf( " Index of Refraction of turbid medium:" );
    scanf( "%lf", &v );  v=3.0e10/v;
    printf( " Mod. Freq. in MHz:" );
    scanf( "%lf", &w );
    w = 2.*pi*w*1.0e6;
  }
  Do = 1/( 3*(us+ua) );
  Zb = 1/us;
  kmag = sqrt( sqrt( (v*v*ua*ua + w*w) / (v*v*Do*Do) ));
  karg = .5 * atan( w/(v*ua) );
  ko = RCmul( kmag, Complex(sin(karg),cos(karg)) );

/*OBJECT INFORMATION*/
  for( i=0; i<nobjs; i++ )
    {
      if( argc==5 )
	fscanf( fp_inphm, "%lf %lf %lf %lf %lf %lf", &usi, &uai[i], &ro[i].x,
	       &ro[i].y, &ro[i].z, &a[i]);
      else {
	printf( "Object %d: mu_s' mu_a\n", i+1 );
	printf( "           Postion Xo Yo Zo\n" );
	printf( "           Size a (units of centimeters):" );
	scanf( "%lf %lf %lf %lf %lf %lf", &usi, &uai[i], &ro[i].x, &ro[i].y, &ro[i].z, &a[i]);
      }

      Di[i] = 1./( 3.*(usi+uai[i]) ); 
      if( uai[i] > AB )  printf( "  This object is being treated as a Perfect Absorber.\n" );
      kmag = sqrt( sqrt( (v*v*uai[i]*uai[i] + w*w) / (v*v*Di[i]*Di[i]) ));
      karg = .5 * atan( w/(v*uai[i]) );
      ki[i] = RCmul( kmag, Complex(sin(karg),cos(karg)) );
    }

/*SOURCE INFORMATION*/
  if( argc!=5 )
    for( i=0; i<nsrcs; i++ ) {
      printf( "SOURCE %d: Position Xs Ys Zs\n", i+1 );
      printf( "           Initial Amplitude and Phase A_o Phi_o:" );
      scanf( "%lf %lf %lf %lf %lf", &rs[i].x, &rs[i].y, &rs[i].z, &foo1, &foo2 );
      rs[i].s.r = foo1 * cos( foo2*pi/180 );
      rs[i].s.i = foo1 * sin( foo2*pi/180 );
    }

/*EXTRAPOLATED ZERO BOUNDARIES*/
  for( i=0; i<nplanes; i++ ) {
    if( argc==5 ) {
      fscanf( fp_inphm, "%lf %lf", &rb[i+1].z, &s );
      rb[i+1].z += Zb*s/fabs(s);
    }
    else {
      printf( "Y of planar boundary %d and direction to air(+1/-1):", i+1 );
      scanf( "%lf %lf", &rb[i+1].y, &s );
      rb[i+1].y += Zb*s/fabs(s);
    }
  }

/*OUTPUT FILE INFORMATION*/
  if( argc!=5 ) {
    printf( "Output filename (excluding extension):" );
    scanf( "%s", filenm );

    printf( "Detector positions (define a 3D grid):\n" );
    printf( " Xmin, Xmax, Xstep:" );
    scanf( "%lf %lf %lf", &xmin, &xmax, &xstep );
    printf( " Ymin, Ymax, Ystep:" );
    scanf( "%lf %lf %lf", &ymin, &ymax, &ystep );
    printf( " Zmin, Zmax, Zstep:" );
    scanf( "%lf %lf %lf", &zmin, &zmax, &zstep );
    nx = (int)( (xmax-xmin)/xstep ) + 1;
    ny = (int)( (ymax-ymin)/ystep ) + 1;
    nz = (int)( (zmax-zmin)/zstep ) + 1;

/*OPEN FILES*/
    sprintf( filetemp, "%s.pdw", filenm );
    fp_outpdw = fopen( filetemp, "w" );
    sprintf( filetemp, "%s.sc", filenm );
    fp_outsc = fopen( filetemp, "w" );
    sprintf( filetemp, "%s.inc", filenm );
    fp_outinc = fopen( filetemp, "w" );
    sprintf( filetemp, "%s.xyz", filenm );
    fp_outxyz = fopen( filetemp, "w" );

/*INITIALIZE DATA FILES*/
    fprintf( fp_outpdw, "%d\n", nx*ny*nz );
    fprintf( fp_outsc, "%d\n", nx*ny*nz );
    fprintf( fp_outinc, "%d\n", nx*ny*nz );
    fprintf( fp_outxyz, "%d\n%d\n%d\n", nx, ny, nz );
  }


/***********************/
/*MAIN CALCULATION LOOP*/
/***********************/
  if( argc==5 ) {
    fscanf( fp_insdg, "%d", &npts );
    fprintf( fp_outpdw, "%d\n", npts );
    fprintf( fp_outsc, "%d\n", npts );
    for( i=0; i<npts; i++ ) {
      for( j=0; j<nsrcs; j++ ) {
	fscanf( fp_insdg, "%lf %lf %lf %lf %lf", &rs[j].x, &rs[j].y, &rs[j].z,
	       &foo1, &foo2 );
	rs[j].s.r = foo1 * cos( foo2*pi/180 );
	rs[j].s.i = foo1 * sin( foo2*pi/180 );
      }
      for( j=0; j<ndets; j++ ) {
	fscanf( fp_insdg, "%lf %lf %lf %lf %lf", &rd[j].x, &rd[j].y, &rd[j].z,
	       &foo1, &foo2 );
	rd[j].s.r = foo1 * cos( foo2*pi/180 );
	rd[j].s.i = foo1 * sin( foo2*pi/180 );
      }
      
      u=DPDW(rd,rs,ro,rb,ko,ki,a,uai,&uinc,&usc,nsrcs,ndets,nobjs,nplanes);

      phase = atan2(u.i,u.r)*180./pi;
      amp = Cabs(u);
      fprintf( fp_outpdw, "%e %lf\n", amp, phase );

      if( amp_o==0. )  amp_o = amp;
      fprintf( fp, "%lf %lf %e %e\n",phase,1.,amp/amp_o,1.);

      phase = atan2(usc.i,usc.r)*180./pi;
      amp = Cabs(usc);
      fprintf( fp_outsc, "%e %lf\n", amp, phase );

      phase = atan2(uinc.i,uinc.r)*180./pi;
      amp = Cabs(uinc);
      fprintf( fp_outinc, "%e %lf\n", amp, phase );
    }
    fclose( fp_insdg );
    fclose( fp_inphm );
    fclose( fp_outpdw );
    fclose( fp_outsc );
    fclose( fp_outinc );
    fclose( fp );
  }
  else {
    rd[0].s = Complex(1.,0.);
    for( k=0; k<nz; k++ ) {
      rd[0].z = zmin + k*zstep;
      for( j=0; j<ny; j++ ) {
	rd[0].y = ymin + j*ystep;
	for( i=0; i<nx; i++ ) {
	  rd[0].x = xmin + i*xstep;
	
	  u=DPDW(rd,rs,ro,rb,ko,ki,a,uai,&uinc,&usc,nsrcs,ndets,nobjs,nplanes);
	
	  if( u.r!=9999.0 ) {
	    if( usc.r!=0. || usc.i!=0 ) {
	      ampsc = Cabs(usc);
	      phsc[i][j] = atan2( usc.i, usc.r );
	      phsc[i][j] *= 180.0 / pi;
	    }
	    else {
	      ampsc = 9999;
	      phsc[i][j] = 9999;
	    }

	    amp = Cabs(u);
	    ph[i][j] = atan2( u.i, u.r );
	    ph[i][j] *= 180.0 / pi;
	  
	    Check( &phsc[i][j], phsc[i-1][j], phsc[i][j-1], &phscold, i, j );
	    Check( &ph[i][j], ph[i-1][j], ph[i][j-1], &phold, i, j );
	  }
	  else {
	    ampsc = 9999.0;  phsc[i][j] = 9999.0;
	    amp = 9999.0;  ph[i][j] = 9999.0;
	  }
	
	  fprintf( fp_outpdw, "%e %lf\n", amp, ph[i][j] );
	  fprintf( fp_outsc, "%e %lf\n", ampsc, phsc[i][j] );

	  amp = Cabs(uinc);
	  phinc[i][j] = atan2(uinc.i,uinc.r)*180./pi;
	  Check( &phinc[i][j], phinc[i-1][j], phinc[i][j-1], &phincold, i, j );
	  fprintf( fp_outinc, "%e %lf\n", amp, phinc[i][j] );

	  fprintf( fp_outxyz, "%lf %lf %lf\n", rd[0].x, rd[0].y, rd[0].z );
	}
      }
    }

/*CLOSE FILES*/
    fclose( fp_outpdw );
    fclose( fp_outsc );
    fclose( fp_outinc );
    fclose( fp_outxyz );
  }

}


/************************************************************
  CHECK THE PHASE ARRAY TO PREVENT 360 DEGREE DISCONTINUITIES
************************************************************/
Check( ph, phx, phz, phold, i, j )
double *ph, phx, phz, *phold;
int i, j;
{
  if( j>0 ) if( (*ph-phz)>180 && phz!=9999.0 ) *ph -= 360;
  if( j>0 ) if( (*ph-phz)<-180 && phz!=9999.0 ) *ph += 360;
  if( i>0 ) if( (*ph-phx)>180 && phx!=9999.0 ) *ph -= 360;
  if( i>0 ) if( (*ph-phx)<-180 && phx!=9999.0 ) *ph += 360;
  if( i>0 && j>0 ) if(phx==9999.0 && phz==9999.0 ) while( (*ph-*phold)>180 ) *ph -= 360;
  if( i>0 && j>0 ) if(phx==9999.0 && phz==9999.0 ) while( (*ph-*phold)<-180 ) *ph += 360;
  *phold = *ph;
}
